# ----------------------------------------------------------------------
#    Copyright (C) 2013 Kshitij Gupta <kgupta8592@gmail.com>
#    Copyright (C) 2015-2024 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------
import os
from shutil import which

import apparmor.aa as apparmor
import apparmor.ui as aaui
from apparmor.common import AppArmorException, cmd, is_skippable_file, user_perm
from apparmor.translations import init_translation

_ = init_translation()


class aa_tools:
    def __init__(self, tool_name, args):
        apparmor.init_aa(profiledir=args.dir, confdir=args.configdir)
        apparmor.read_profiles(skip_disabled=(tool_name != 'enforce'))

        if not user_perm(apparmor.profile_dir):
            raise AppArmorException("Cannot write to profile directory: %s" % (apparmor.profile_dir))

        self.name = tool_name
        self.profiling = args.program
        self.silent = None
        self.do_reload = args.do_reload
        self.requires_binary = tool_name in ('autodep',)

        self.explicit_profile = getattr(args, 'profile', None)
        self.explicit_profile_file = getattr(args, 'profile_file', None)
        self.explicit_executable = getattr(args, 'executable_file', None)

        if tool_name == 'audit':
            self.remove = args.remove
        elif tool_name == 'autodep':
            self.force = args.force
            self.aa_mountpoint = apparmor.check_for_apparmor()
        elif tool_name == 'cleanprof':
            self.silent = args.silent

    def _resolve_path_argument(self, p):
        """Resolve a path argument to (program, profile, prof_filename) or None if invalid."""
        fq_path = apparmor.get_full_path(p).strip()

        if os.path.commonprefix([apparmor.profile_dir, fq_path]) == apparmor.profile_dir:
            return (None, None, fq_path)

        program = fq_path
        profile = apparmor.active_profiles.profile_from_attachment(fq_path)
        prof_filename = apparmor.get_profile_filename_from_attachment(fq_path, True)

        if not os.path.exists(fq_path) and not self.requires_binary:
            has_profile = profile and apparmor.active_profiles.profile_exists(profile)
            has_file = os.path.isfile(prof_filename)
            if not has_profile and not has_file:
                aaui.UI_Info(_("%s does not exist, please double-check the path.") % p)
                return None

        return (program, profile, prof_filename)

    def _resolve_name_argument(self, p):
        """Resolve a name argument to (program, profile, prof_filename) or None if invalid."""
        if apparmor.active_profiles.profile_exists(p):
            prof_filename = apparmor.get_profile_filename_from_profile_name(p)
            return (p, p, prof_filename)

        which_ = which(p)
        if which_ is not None:
            program = apparmor.get_full_path(which_)
            prof_filename = apparmor.get_profile_filename_from_attachment(program, True)
            return (program, program, prof_filename)

        profile_path = os.path.join(apparmor.profile_dir, p)
        if os.path.exists(profile_path):
            prof_filename = apparmor.get_full_path(profile_path).strip()
            return (None, p, prof_filename)

        if '/' not in p:
            aaui.UI_Info(_("Can't find %(program)s in the system path list. If the name of the application\nis correct, please run 'which %(program)s' as a user with correct PATH\nenvironment set up in order to find the fully-qualified path and\nuse the full path as parameter.")
                         % {'program': p})
        else:
            aaui.UI_Info(_("%s does not exist, please double-check the path.") % p)

        return None

    def get_next_to_profile(self):
        """Iterator function to walk the list of arguments passed"""

        if self.explicit_profile_file:
            prof_filename = apparmor.get_full_path(self.explicit_profile_file).strip()
            if not os.path.isfile(prof_filename):
                aaui.UI_Info(_("Profile file %s does not exist.") % self.explicit_profile_file)
            else:
                profiles = list(apparmor.active_profiles.profiles_in_file(prof_filename))
                if not profiles:
                    aaui.UI_Info(_("No profiles found in %s.") % prof_filename)
                else:
                    for profile in profiles:
                        yield (profile, profile, prof_filename)

        if self.explicit_profile:
            profile = self.explicit_profile
            if apparmor.active_profiles.profile_exists(profile):
                prof_filename = apparmor.get_profile_filename_from_profile_name(profile)
                yield (profile, profile, prof_filename)
            elif os.path.exists(os.path.join(apparmor.profile_dir, profile)):
                prof_filename = apparmor.get_full_path(os.path.join(apparmor.profile_dir, profile)).strip()
                yield (None, profile, prof_filename)
            else:
                aaui.UI_Info(_("Profile for %s not found, skipping") % profile)

        if self.explicit_executable:
            program = apparmor.get_full_path(self.explicit_executable).strip()
            profile = apparmor.active_profiles.profile_from_attachment(program)
            prof_filename = apparmor.get_profile_filename_from_attachment(program, True)
            if not profile or not apparmor.active_profiles.profile_exists(profile):
                if not os.path.isfile(prof_filename):
                    aaui.UI_Info(_("Profile for %s not found, skipping") % program)
                else:
                    profiles = list(apparmor.active_profiles.profiles_in_file(prof_filename))
                    if profiles:
                        for prof in profiles:
                            yield (program, prof, prof_filename)
                    else:
                        aaui.UI_Info(_("Profile for %s not found, skipping") % program)
            else:
                yield (program, profile, prof_filename)

        for p in self.profiling:
            if not p:
                continue

            if os.path.exists(p) or p.startswith('/'):
                result = self._resolve_path_argument(p)
            else:
                result = self._resolve_name_argument(p)

            if result:
                yield result

    def get_next_for_modechange(self):
        """common code for mode/flags changes"""

        for (program, _ignored, prof_filename) in self.get_next_to_profile():
            output_name = prof_filename if program is None else program

            if not os.path.isfile(prof_filename) or is_skippable_file(prof_filename):
                aaui.UI_Info(_('Profile for %s not found, skipping') % output_name)
                continue

            yield (program, prof_filename, output_name)

    def _clean_profiles_in_file(self, prof_filename):
        """Clean all profiles found in the given file. Returns True if profiles were found and cleaned."""
        if prof_filename and os.path.isfile(prof_filename):
            profiles_in_file = list(apparmor.active_profiles.profiles_in_file(prof_filename))
            if profiles_in_file:
                for prof in profiles_in_file:
                    self.clean_profile(prof, prof, prof_filename)
                return True
        return False

    def cleanprof_act(self):
        for (program, profile, prof_filename) in self.get_next_to_profile():
            if program is None:
                program = profile

            if profile and apparmor.active_profiles.profile_exists(profile):
                self.clean_profile(program, profile, prof_filename)
            elif self._clean_profiles_in_file(prof_filename):
                pass
            else:
                output_name = program if program else prof_filename
                aaui.UI_Info(_('Profile for %s not found, skipping') % output_name)

    def cmd_disable(self):
        for (program, prof_filename, output_name) in self.get_next_for_modechange():
            aaui.UI_Info(_('Disabling %s.') % output_name)

            apparmor.create_symlink('disable', prof_filename)

            self.unload_profile(prof_filename)

    def cmd_enforce(self):
        for (program, prof_filename, output_name) in self.get_next_for_modechange():
            apparmor.set_enforce(prof_filename, program)

            self.reload_profile(prof_filename)

    def cmd_complain(self):
        for (program, prof_filename, output_name) in self.get_next_for_modechange():
            apparmor.set_complain(prof_filename, program)

            self.reload_profile(prof_filename)

    def cmd_audit(self):
        for (program, prof_filename, output_name) in self.get_next_for_modechange():

            # keep this to allow toggling 'audit' flags
            if not self.remove:
                aaui.UI_Info(_('Setting %s to audit mode.') % output_name)
            else:
                aaui.UI_Info(_('Removing audit mode from %s.') % output_name)
            apparmor.change_profile_flags(prof_filename, program, 'audit', not self.remove)

            disable_link = '%s/disable/%s' % (apparmor.profile_dir, os.path.basename(prof_filename))

            if os.path.exists(disable_link):
                aaui.UI_Info(_('\nWarning: the profile %s is disabled. Use aa-enforce or aa-complain to enable it.') % os.path.basename(prof_filename))

            self.reload_profile(prof_filename)

    def cmd_autodep(self):
        apparmor.loadincludes()

        for (program, _ignored, prof_filename) in self.get_next_to_profile():
            if not program:
                aaui.UI_Info(_('Please pass an application to generate a profile for, not a profile itself - skipping %s.') % prof_filename)
                continue

            apparmor.check_qualifiers(program)

            if os.path.exists(apparmor.get_profile_filename_from_attachment(program, True)) and not self.force:
                aaui.UI_Info(_('Profile for %s already exists - skipping.') % program)
            else:
                apparmor.autodep(program)
                if self.aa_mountpoint:
                    apparmor.reload(program)

    def clean_profile(self, program, profile, prof_filename):
        import apparmor.cleanprofile as cleanprofile

        if not prof_filename:
            raise AppArmorException(_('The profile for %s does not exist. Nothing to clean.') % program)

        prof = cleanprofile.Prof(prof_filename)
        cleanprof = cleanprofile.CleanProf(True, prof, prof)
        deleted = cleanprof.remove_duplicate_rules(profile)
        aaui.UI_Info(_("\nDeleted %s rules.") % deleted)
        apparmor.changed[profile] = True

        if self.silent:
            apparmor.write_profile_ui_feedback(profile, True)
            self.reload_profile(prof_filename)
        else:
            q = aaui.PromptQuestion()
            q.title = 'Changed Local Profiles'
            q.explanation = _('The local profile for %(program)s in file %(file)s was changed. Would you like to save it?') % {'program': program, 'file': prof_filename}
            q.functions = ['CMD_SAVE_CHANGES', 'CMD_VIEW_CHANGES', 'CMD_ABORT']
            q.default = 'CMD_VIEW_CHANGES'
            q.options = []
            q.selected = 0
            ans = ''
            arg = None
            while ans != 'CMD_SAVE_CHANGES':
                ans, arg = q.promptUser()
                if ans == 'CMD_SAVE_CHANGES':
                    apparmor.write_profile_ui_feedback(profile)
                    self.reload_profile(prof_filename)
                elif ans == 'CMD_VIEW_CHANGES':
                    # oldprofile = apparmor.serialize_profile(apparmor.original_profiles, profile, {})
                    newprofile = apparmor.serialize_profile(apparmor.active_profiles, profile, {})  # , {'is_attachment': True})
                    aaui.UI_Changes(prof_filename, newprofile, comments=True)

    def unload_profile(self, prof_filename):
        if not self.do_reload:
            return

        # FIXME: should ensure profile is loaded before unloading
        cmd_info = cmd([apparmor.parser, '-I%s' % apparmor.profile_dir, '--base', apparmor.profile_dir, '-R', prof_filename])

        if cmd_info[0] != 0:
            raise AppArmorException(cmd_info[1])

    def reload_profile(self, prof_filename):
        if not self.do_reload:
            return

        apparmor.reload_profile(prof_filename, raise_exc=True)



# ----------------------------------------------------------------------
#    Copyright (C) 2015 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

import re

from apparmor.common import AppArmorBug, AppArmorException
from apparmor.regex import RE_PROFILE_DBUS, strip_parenthesis, strip_quotes, re_cond
from apparmor.rule import (
    BaseRule, BaseRuleset, check_and_split_list, logprof_value_or_all,
    parse_modifiers, quote_if_needed)
from apparmor.translations import init_translation

_ = init_translation()

message_keywords = ['send', 'receive', 'r', 'read', 'w', 'write', 'rw']
access_keywords = ['bind', 'eavesdrop'] + message_keywords

# XXX joint_access_keyword and RE_ACCESS_KEYWORDS exactly as in SignalRule - move to function?
joint_access_keyword = '(' + r'(\s|,)*' + '(' + '|'.join(access_keywords) + r')(\s|,)*' + ')'
RE_ACCESS_KEYWORDS = (
    joint_access_keyword  # one of the access_keyword
    + '|'  # or
    + r'\(' + r'(\s|,)*' + joint_access_keyword + '?' + '(' + r'(\s|,)+' + joint_access_keyword + ')*' + r'\)'  # one or more access_keyword in (...)
)


RE_FLAG = r'(?P<%s>(\S+|"[^"]+"|\(\s*\S+\s*\)|\(\s*"[^"]+"\)\s*))'  # string without spaces, or quoted string, optionally wrapped in (...). %s is the match group name
# plaintext version:      | * | "* "  | (    *    ) | (  " *   " )    |

# XXX this regex will allow repeated parameters, last one wins
# XXX (the parser will reject such rules)
RE_DBUS_DETAILS = re.compile(
    '^'
    + r'(\s+(?P<access>' + RE_ACCESS_KEYWORDS + '))?'  # optional access keyword(s)
        + '('  # noqa: E131
            + r'(\s+(bus\s*=\s*'       + RE_FLAG % 'bus'       + '))?|'  # optional bus= system | session | AARE, (...) optional  # noqa: E131,E221
            + r'(\s+(path\s*=\s*'      + RE_FLAG % 'path'      + '))?|'  # optional path=AARE, (...) optional  # noqa: E221
            + r'(\s+(name\s*=\s*'      + RE_FLAG % 'name'      + '))?|'  # optional name=AARE, (...) optional  # noqa: E221
            + r'(\s+(interface\s*=\s*' + RE_FLAG % 'interface' + '))?|'  # optional interface=AARE, (...) optional
            + r'(\s+(member\s*=\s*'    + RE_FLAG % 'member'    + '))?|'  # optional member=AARE, (...) optional  # noqa: E221
            + r'(\s+(peer\s*=\s*\(([,\s]*(' + re_cond("name") + '|' + re_cond("label") + r')(\s*,)?)*\)))?'  # noqa: E221
        + '){0,6}'
    + r'\s*$')


class DbusRule(BaseRule):
    """Class to handle and store a single dbus rule"""

    # Nothing external should reference this class, all external users
    # should reference the class field DbusRule.ALL
    class __DbusAll:
        pass

    ALL = __DbusAll

    rule_name = 'dbus'
    _match_re = RE_PROFILE_DBUS

    def __init__(self, access, bus, path, name, interface, member, peername, peerlabel,
                 audit=False, deny=False, allow_keyword=False, comment='', log_event=None, priority=None):

        super().__init__(audit=audit, deny=deny, allow_keyword=allow_keyword,
                         comment=comment, log_event=log_event, priority=priority)

        self.access, self.all_access, unknown_items = check_and_split_list(access, access_keywords, self.ALL, type(self).__name__, 'access')
        if unknown_items:
            raise AppArmorException(_('Passed unknown access keyword to %s: %s') % (type(self).__name__, ' '.join(unknown_items)))

        #                                                       rulepart   partname    is_path  log_event
        self.bus, self.all_buses            = self._aare_or_all(bus,       'bus',        False, log_event)  # noqa: E221
        self.path, self.all_paths           = self._aare_or_all(path,      'path',       True,  log_event)  # noqa: E221
        self.name, self.all_names           = self._aare_or_all(name,      'name',       False, log_event)  # noqa: E221
        self.interface, self.all_interfaces = self._aare_or_all(interface, 'interface',  False, log_event)  # noqa: E221
        self.member, self.all_members       = self._aare_or_all(member,    'member',     False, log_event)  # noqa: E221
        self.peername, self.all_peernames   = self._aare_or_all(peername,  'peer name',  False, log_event)  # noqa: E221
        self.peerlabel, self.all_peerlabels = self._aare_or_all(peerlabel, 'peer label', False, log_event)  # noqa: E221

        # not all combinations are allowed
        if self.access and 'bind' in self.access and (self.path or self.interface or self.member or self.peername or self.peerlabel):
            raise AppArmorException(_('dbus bind rules must not contain a path, interface, member or peer conditional'))
        elif self.access and 'eavesdrop' in self.access and (self.name or self.path or self.interface or self.member or self.peername or self.peerlabel):
            raise AppArmorException(_('dbus eavesdrop rules must not contain a name, path, interface, member or peer conditional'))
        elif self.access and self.name:
            for msg in message_keywords:
                if msg in self.access:
                    raise AppArmorException(_('dbus %s rules must not contain a name conditional') % '/'.join(self.access))

    @classmethod
    def _create_instance(cls, raw_rule, matches):
        """parse raw_rule and return instance of this class"""

        priority, audit, deny, allow_keyword, comment = parse_modifiers(matches)

        rule_details = ''
        if matches.group('details'):
            rule_details = matches.group('details')

        if rule_details:
            details = RE_DBUS_DETAILS.search(rule_details)
            if not details:
                raise AppArmorException(_("Invalid or unknown keywords in 'dbus %s" % rule_details))

            if details.group('access'):
                # XXX move to function _split_access()?
                access = strip_parenthesis(details.group('access'))
                access = access.replace(',', ' ').split()  # split by ',' or whitespace
                if not access:  # XXX that happens for "dbus ( )," rules - correct behaviour? (also: same for signal rules?)
                    access = cls.ALL
            else:
                access = cls.ALL

            if details.group('bus'):
                bus = strip_parenthesis(strip_quotes(details.group('bus')))
            else:
                bus = cls.ALL

            if details.group('path'):
                path = strip_parenthesis(strip_quotes(details.group('path')))
            else:
                path = cls.ALL

            if details.group('name'):
                name = strip_parenthesis(strip_quotes(details.group('name')))
            else:
                name = cls.ALL

            if details.group('interface'):
                interface = strip_parenthesis(strip_quotes(details.group('interface')))
            else:
                interface = cls.ALL

            if details.group('member'):
                member = strip_parenthesis(strip_quotes(details.group('member')))
            else:
                member = cls.ALL

            if details.group('name_cond'):
                peername = strip_parenthesis(strip_quotes(details.group('name_cond')))
            else:
                peername = cls.ALL

            if details.group('label_cond'):
                peerlabel = strip_parenthesis(strip_quotes(details.group('label_cond')))
            else:
                peerlabel = cls.ALL

        else:
            access = cls.ALL
            bus = cls.ALL
            path = cls.ALL
            name = cls.ALL
            interface = cls.ALL
            member = cls.ALL
            peername = cls.ALL
            peerlabel = cls.ALL

        return cls(access, bus, path, name, interface, member, peername, peerlabel,
                   audit=audit, deny=deny, allow_keyword=allow_keyword, comment=comment,
                   priority=priority)

    def get_clean(self, depth=0):
        """return rule (in clean/default formatting)"""

        space = '  ' * depth

        # XXX split off _get_access_rule_part? (also needed in PtraceRule)
        if self.all_access:
            access = ''
        elif len(self.access) == 1:
            access = ' %s' % ' '.join(self.access)
        elif self.access:
            access = ' (%s)' % ' '.join(sorted(self.access))
        else:
            raise AppArmorBug('Empty access in dbus rule')

        bus         = self._get_aare_rule_part('bus',       self.bus,       self.all_buses)  # noqa: E221
        path        = self._get_aare_rule_part('path',      self.path,      self.all_paths)  # noqa: E221
        name        = self._get_aare_rule_part('name',      self.name,      self.all_names)  # noqa: E221
        interface   = self._get_aare_rule_part('interface', self.interface, self.all_interfaces)  # noqa: E221
        member      = self._get_aare_rule_part('member',    self.member,    self.all_members)  # noqa: E221

        peername    = self._get_aare_rule_part('name',      self.peername,  self.all_peernames)  # noqa: E221
        peerlabel   = self._get_aare_rule_part('label',     self.peerlabel, self.all_peerlabels)  # noqa: E221
        peer = peername + peerlabel
        if peer:
            peer = ' peer=(%s)' % peer.strip()

        return ('%s%sdbus%s%s%s%s%s%s%s,%s' % (space, self.modifiers_str(), access, bus, path, name, interface, member, peer, self.comment))

    def _get_aare_rule_part(self, prefix, value, all_values):
        """helper function to write a rule part
           value is expected to be a AARE"""
        if all_values:
            return ''
        elif value:
            if re.match("(?<!@){", value.regex):  # apparmor_parser doesn't properly handle unquoted {} AARE
                return ' %(prefix)s="%(value)s"' % {'prefix': prefix, 'value': value.regex}
            else:
                return ' %(prefix)s=%(value)s' % {'prefix': prefix, 'value': quote_if_needed(value.regex)}
        else:
            raise AppArmorBug('Empty %(prefix_name)s in %(rule_name)s rule' % {'prefix_name': prefix, 'rule_name': self.rule_name})

    def _is_covered_localvars(self, other_rule):
        """check if other_rule is covered by this rule object"""

        if not self._is_covered_list(self.access,       self.all_access,        other_rule.access,      other_rule.all_access,      'access'):
            return False

        if not self._is_covered_aare(self.bus,          self.all_buses,         other_rule.bus,         other_rule.all_buses,       'bus'):
            return False

        if not self._is_covered_aare(self.path,         self.all_paths,         other_rule.path,        other_rule.all_paths,       'path'):
            return False

        if not self._is_covered_aare(self.name,         self.all_names,         other_rule.name,        other_rule.all_names,       'name'):
            return False

        if not self._is_covered_aare(self.interface,    self.all_interfaces,    other_rule.interface,   other_rule.all_interfaces,  'interface'):
            return False

        if not self._is_covered_aare(self.member,       self.all_members,       other_rule.member,      other_rule.all_members,     'member'):
            return False

        if not self._is_covered_aare(self.peername,     self.all_peernames,     other_rule.peername,    other_rule.all_peernames,   'peername'):
            return False

        if not self._is_covered_aare(self.peerlabel,    self.all_peerlabels,    other_rule.peerlabel,   other_rule.all_peerlabels,  'peerlabel'):
            return False

        # still here? -> then it is covered
        return True

    def _is_equal_localvars(self, other_rule, strict):
        """compare if rule-specific variables are equal"""

        if (self.access != other_rule.access
                or self.all_access != other_rule.all_access):
            return False

        if not self._is_equal_aare(self.bus,        self.all_buses,         other_rule.bus,         other_rule.all_buses,         'bus'):
            return False

        if not self._is_equal_aare(self.path,       self.all_paths,         other_rule.path,        other_rule.all_paths,         'path'):
            return False

        if not self._is_equal_aare(self.name,       self.all_names,         other_rule.name,        other_rule.all_names,         'name'):
            return False

        if not self._is_equal_aare(self.interface,  self.all_interfaces,    other_rule.interface,   other_rule.all_interfaces,    'interface'):
            return False

        if not self._is_equal_aare(self.member,     self.all_members,       other_rule.member,      other_rule.all_members,       'member'):
            return False

        if not self._is_equal_aare(self.peername,   self.all_peernames,     other_rule.peername,    other_rule.all_peernames,     'peername'):
            return False

        if not self._is_equal_aare(self.peerlabel,  self.all_peerlabels,    other_rule.peerlabel,   other_rule.all_peerlabels,    'peerlabel'):
            return False

        return True

    def _logprof_header_localvars(self):
        access    = logprof_value_or_all(self.access,    self.all_access)     # noqa: E221
        bus       = logprof_value_or_all(self.bus,       self.all_buses)      # noqa: E221
        path      = logprof_value_or_all(self.path,      self.all_paths)      # noqa: E221
        name      = logprof_value_or_all(self.name,      self.all_names)      # noqa: E221
        interface = logprof_value_or_all(self.interface, self.all_interfaces)
        member    = logprof_value_or_all(self.member,    self.all_members)    # noqa: E221
        peername  = logprof_value_or_all(self.peername,  self.all_peernames)  # noqa: E221
        peerlabel = logprof_value_or_all(self.peerlabel, self.all_peerlabels)

        return (
            _('Access mode'), access,
            _('Bus'),         bus,
            _('Path'),        path,
            _('Name'),        name,
            _('Interface'),   interface,
            _('Member'),      member,
            _('Peer name'),   peername,
            _('Peer label'),  peerlabel,
        )

    @staticmethod
    def hashlog_from_event(hl, ev):
        hl[ev['denied_mask']][ev['bus']][ev['path']][ev['name']][ev['interface']][ev['member']][ev['peer_profile']] = True

    @classmethod
    def from_hashlog(cls, hl):
        for access, bus, path, name, interface, member, peer_profile in BaseRule.generate_rules_from_hashlog(hl, 7):
            # Depending on the access type, not all parameters are allowed.
            # Ignore them, even if some of them appear in the log.
            # Also, the log doesn't provide a peer name, therefore always use ALL.
            if access in ('send', 'receive'):
                if peer_profile is None:
                    peer_profile = cls.ALL
                yield cls(access, bus, path, cls.ALL, interface, member, cls.ALL, peer_profile, log_event=True)
            elif access == 'bind':
                yield cls(access, bus, cls.ALL, name, cls.ALL, cls.ALL, cls.ALL, cls.ALL, log_event=True)
            elif access == 'eavesdrop':
                yield cls(access, bus, cls.ALL, cls.ALL, cls.ALL, cls.ALL, cls.ALL, cls.ALL, log_event=True)
            else:
                raise AppArmorBug('unexpected dbus access: {}'.format(access))


class DbusRuleset(BaseRuleset):
    """Class to handle and store a collection of dbus rules"""

    def get_glob(self, path_or_rule):
        """Return the next possible glob. For dbus rules, that means removing access or removing/globbing bus"""
        # XXX only remove one part, not all
        return 'dbus,'



# ----------------------------------------------------------------------
#    Copyright (C) 2020 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

from apparmor.common import AppArmorBug, AppArmorException
from apparmor.regex import RE_PROFILE_ALIAS, strip_quotes
from apparmor.rule import BaseRule, BaseRuleset, parse_comment, quote_if_needed
from apparmor.translations import init_translation

_ = init_translation()


class AliasRule(BaseRule):
    """Class to handle and store a single alias rule"""

    rule_name = 'alias'
    _match_re = RE_PROFILE_ALIAS

    def __init__(self, orig_path, target, audit=False, deny=False, allow_keyword=False,
                 comment='', log_event=None, priority=None):

        super().__init__(audit=audit, deny=deny, allow_keyword=allow_keyword,
                         comment=comment, log_event=log_event, priority=priority)

        # aliases don't support priority, allow keyword, audit or deny
        self.ensure_modifiers_not_supported()

        if not isinstance(orig_path, str):
            raise AppArmorBug('Passed unknown type for orig_path to %s: %s' % (self.__class__.__name__, orig_path))
        if not orig_path:
            raise AppArmorException('Passed empty orig_path to %s: %s' % (self.__class__.__name__, orig_path))
        if not orig_path.startswith('/'):
            raise AppArmorException("Alias path doesn't start with '/'")

        if not isinstance(target, str):
            raise AppArmorBug('Passed unknown type for target to %s: %s' % (self.__class__.__name__, target))
        if not target:
            raise AppArmorException('Passed empty target to %s: %s' % (self.__class__.__name__, target))
        if not target.startswith('/'):
            raise AppArmorException("Alias target doesn't start with '/'")

        self.orig_path = orig_path
        self.target = target

    @classmethod
    def _create_instance(cls, raw_rule, matches):
        """parse raw_rule and return instance of this class"""

        comment = parse_comment(matches)

        orig_path = strip_quotes(matches.group('orig_path').strip())
        target = strip_quotes(matches.group('target').strip())

        return cls(orig_path, target,
                   audit=False, deny=False, allow_keyword=False, comment=comment, priority=None)

    def get_clean(self, depth=0):
        """return rule (in clean/default formatting)"""

        space = '  ' * depth

        return '%salias %s -> %s,' % (space, quote_if_needed(self.orig_path), quote_if_needed(self.target))

    def _is_covered_localvars(self, other_rule):
        """check if other_rule is covered by this rule object"""

        # the only way aliases can be covered are exact duplicates
        return self._is_equal_localvars(other_rule, False)

    def _is_equal_localvars(self, other_rule, strict):
        """compare if rule-specific aliases are equal"""

        if self.orig_path != other_rule.orig_path:
            return False

        if self.target != other_rule.target:
            return False

        return True

    def _logprof_header_localvars(self):
        return _('Alias'), '%s -> %s' % (self.orig_path, self.target)


class AliasRuleset(BaseRuleset):
    """Class to handle and store a collection of alias rules"""



# ----------------------------------------------------------------------
#    Copyright (C) 2013 Kshitij Gupta <kgupta8592@gmail.com>
#    Copyright (C) 2014-2015 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

import re
import itertools

from apparmor.common import AppArmorBug, AppArmorException
from apparmor.translations import init_translation

_ = init_translation()

# Profile parsing Regex
RE_PRIORITY_AUDIT_DENY = r'^\s*(priority\s*=\s*(?P<priority>[+-]?[0-9]*)\s+)?(?P<audit>audit\s+)?(?P<allow>allow\s+|deny\s+)?'  # line start, optionally: leading whitespace, <priority> = number, <audit> and <allow>/deny
RE_EOL = r'\s*(?P<comment>#.*?)?\s*$'  # optional whitespace, optional <comment>, optional whitespace, end of the line
RE_COMMA_EOL = r'\s*,' + RE_EOL  # optional whitespace, comma + RE_EOL

RE_PROFILE_NAME = r'(?P<%s>(\S+|"[^"]+"))'  # string without spaces, or quoted string. %s is the match group name
RE_PATH = r'/\S*|"/[^"]*"'  # filename (starting with '/') without spaces, or quoted filename.
RE_VAR = r'@{[^,}\s]+}'
RE_DICT_ENTRY = r'\s*(?P<key>[^,\s=]+)(?:=(?P<value>[^,\s=]+))?\s*'
RE_PROFILE_PATH = '(?P<%s>(' + RE_PATH + '))'  # quoted or unquoted filename. %s is the match group name
RE_PROFILE_PATH_OR_VAR = '(?P<%s>(' + RE_PATH + '|' + RE_VAR + r'\S*|"' + RE_VAR + '[^"]*"))'  # quoted or unquoted filename or variable. %s is the match group name
RE_SAFE_OR_UNSAFE = '(?P<execmode>(safe|unsafe))'
RE_XATTRS = r'(\s+xattrs\s*=\s*\((?P<xattrs>([^)=]+(=[^)=]+)?\s?)*)\)\s*)?'
RE_FLAGS = r'(\s+(flags\s*=\s*)?\((?P<flags>[^)]+)\))?'

RE_VARIABLE = re.compile(RE_VAR)
RE_ID = r'(?P<id%(label)s>[^,!#\s=@$()"]+|"(\w|\s)*")'
RE_VARIABLES = r'(?P<var%(label)s>(?P<var_type%(label)s>@|\$)\{?(?P<varname%(label)s>\w+)\}?)'
RE_ID_OR_VAR = r'(' + RE_VARIABLES + r'|' + RE_ID + r')'
RE_ALL_VARIABLES = re.compile(RE_VARIABLES % {'label': ''})

RE_PROFILE_END = re.compile(r'^\s*\}' + RE_EOL)
RE_PROFILE_ALL = re.compile(RE_PRIORITY_AUDIT_DENY + r'all' + RE_COMMA_EOL)
RE_PROFILE_CAP = re.compile(RE_PRIORITY_AUDIT_DENY + r'capability(?P<capability>(\s+\S+)+)?' + RE_COMMA_EOL)
RE_PROFILE_ALIAS = re.compile(r'^\s*alias\s+(?P<orig_path>"??.+?"??)\s+->\s*(?P<target>"??.+?"??)' + RE_COMMA_EOL)
RE_PROFILE_RLIMIT = re.compile(r'^\s*set\s+rlimit\s+(?P<rlimit>[a-z]+)\s*<=\s*(?P<value>[^ ]+(\s+[a-zA-Z]+)?)' + RE_COMMA_EOL)
RE_PROFILE_BOOLEAN = re.compile(r'^\s*(?P<varname>\$\{?\w*\}?)\s*=\s*(?P<value>true|false)\s*,?' + RE_EOL, flags=re.IGNORECASE)
RE_PROFILE_VARIABLE = re.compile(r'^\s*(?P<varname>@\{?\w+\}?)\s*(?P<mode>\+?=)\s*(?P<values>@*.+?)' + RE_EOL)

RE_BOOLEAN_OP = r'(?P<boolean_op%(term)s>(?P<boolean_not%(term)s>(not\s+)*)(?P<defined%(term)s>defined\s+)?' + RE_VARIABLES % {'label': '%(term)s'} + r')'
RE_COMPARE_OP_QUOTED = r'(?P<compare_op%(term)s>(?P<compare_not%(term)s>(not\s+)*)(?P<left%(term)s>"?' + RE_ID_OR_VAR % {'label': '_left%(term)s'} + r'"?)\s+(?P<op%(term)s>==|!=|in|>|>=|<|<=)\s+(?P<right%(term)s>"?' + RE_ID_OR_VAR % {'label': '_right%(term)s'} + r'"?))'  # used only by transform_cond
RE_COMPARE_OP = RE_COMPARE_OP_QUOTED.replace('"?', '')

RE_FACTOR = r'(?P<open_paren%(term)s>\()?(' + RE_COMPARE_OP + r'|' + RE_BOOLEAN_OP + r')(?P<close_paren%(term)s>\))?'

RE_TERM = r'(?P<open_paren%(expr)s>\()*\s*((?P<one%(expr)s>' + RE_FACTOR % {'term': '_1%(expr)s'} + r')\s+(?P<cond_op%(expr)s>and|or)\s+(?P<two%(expr)s>' + RE_FACTOR % {'term': '_2%(expr)s'} + ')|' + RE_FACTOR % {'term': '_0%(expr)s'} + r')\s*(?P<close_paren%(expr)s>\))*'

RE_CONDITION = r'(?P<expr>(?P<open_paren>\()?(?P<first>' + RE_TERM % {'expr': '_first'} + r')(\s+(?P<cond_op>and|or)\s+(?P<second>' + RE_TERM % {'expr': '_second'} + r'))*(?P<close_paren>\))?)'

RE_PROFILE_CONDITIONAL = r'\s*if\s+' + RE_CONDITION + r'\s*\{'

RE_PROFILE_CONDITIONAL_START = re.compile(r'^' + RE_PROFILE_CONDITIONAL + RE_EOL)
RE_PROFILE_CONDITIONAL_ELSE = re.compile(r'^\s*(?P<close>\})?\s*else((?P<if>\s+' + RE_PROFILE_CONDITIONAL + r')|(\s*\{))' + RE_EOL)
RE_PROFILE_NETWORK = re.compile(RE_PRIORITY_AUDIT_DENY + r'network(?P<details>\s+.*)?' + RE_COMMA_EOL)
RE_PROFILE_CHANGE_HAT = re.compile(r'^\s*\^("??.+?"??)' + RE_COMMA_EOL)
RE_PROFILE_HAT_DEF = re.compile(r'^(?P<leadingspace>\s*)(?P<hat_keyword>\^|hat\s+)(?P<hat>"??[^)]+?"??)' + RE_FLAGS + r'\s*\{' + RE_EOL)
RE_PROFILE_DBUS = re.compile(RE_PRIORITY_AUDIT_DENY + r'(dbus\s*,|dbus(?P<details>\s+[^#]*)\s*,)' + RE_EOL)
RE_PROFILE_MOUNT = re.compile(RE_PRIORITY_AUDIT_DENY + r'((?P<operation>mount|remount|umount|unmount)(?P<details>\s+[^#]*)?\s*,)' + RE_EOL)
RE_PROFILE_SIGNAL = re.compile(RE_PRIORITY_AUDIT_DENY + r'(signal\s*,|signal(?P<details>\s+[^#]*)\s*,)' + RE_EOL)
RE_PROFILE_PTRACE = re.compile(RE_PRIORITY_AUDIT_DENY + r'(ptrace\s*,|ptrace(?P<details>\s+[^#]*)\s*,)' + RE_EOL)
RE_PROFILE_PIVOT_ROOT = re.compile(RE_PRIORITY_AUDIT_DENY + r'(pivot_root\s*,|pivot_root(?P<details>\s+[^#]*),)' + RE_EOL)
RE_PROFILE_UNIX = re.compile(RE_PRIORITY_AUDIT_DENY + r'(unix\s*,|unix(?P<details>\s+[^#]*)\s*,)' + RE_EOL)
RE_PROFILE_USERNS = re.compile(RE_PRIORITY_AUDIT_DENY + r'(userns\s*,|userns(?P<details>\s+[^#]*)\s*,)' + RE_EOL)
RE_PROFILE_MQUEUE = re.compile(RE_PRIORITY_AUDIT_DENY + r'(mqueue\s*,|mqueue(?P<details>\s+[^#]*)\s*,)' + RE_EOL)
RE_PROFILE_IO_URING = re.compile(RE_PRIORITY_AUDIT_DENY + r'(io_uring\s*,|io_uring(?P<details>\s+[^#]*)\s*,)' + RE_EOL)

RE_METADATA_LOGPROF_SUGGEST = re.compile(r'^\s*#\s*LOGPROF-SUGGEST\s*:\s*(?P<suggest>.*)$')

# match anything that's not " or #, or matching quotes with anything except quotes inside
__re_no_or_quoted_hash = '([^#"]|"[^"]*")*'

RE_RULE_HAS_COMMA = re.compile(
    '^' + __re_no_or_quoted_hash
    + r',\s*(#.*)?$')  # match comma plus any trailing comment
RE_HAS_COMMENT_SPLIT = re.compile(
    '^(?P<not_comment>' + __re_no_or_quoted_hash + ')'  # store in 'not_comment' group
    + '(?P<comment>#.*)$')  # match trailing comment and store in 'comment' group


RE_PROFILE_START = re.compile(
    r'^(?P<leadingspace>\s*)'
    + '('
        + RE_PROFILE_PATH_OR_VAR % 'plainprofile'  # just a path # noqa: E131
        + '|'  # or
        + '(' + 'profile' + r'\s+' + RE_PROFILE_NAME % 'namedprofile' + r'(\s+' + RE_PROFILE_PATH_OR_VAR % 'attachment' + ')?' + ')'  # 'profile', profile name, optionally attachment
    + ')'
    + RE_XATTRS
    + RE_FLAGS
    + r'\s*\{'
    + RE_EOL)


RE_PROFILE_CHANGE_PROFILE = re.compile(
    RE_PRIORITY_AUDIT_DENY
    + 'change_profile'
    + r'(\s+' + RE_SAFE_OR_UNSAFE + ')?'  # optionally exec mode
    + r'(\s+' + RE_PROFILE_PATH_OR_VAR % 'execcond' + ')?'  # optionally exec condition
    + r'(\s+->\s*' + RE_PROFILE_NAME % 'targetprofile' + ')?'  # optionally '->' target profile
    + RE_COMMA_EOL)


# RE_PATH_PERMS is as restrictive as possible, but might still cause mismatches when adding different rule types.
# Therefore parsing code should match against file rules only after trying to match all other rule types.
RE_PATH_PERMS = '(?P<%s>[mrwalkPUCpucix]+)'

RE_PROFILE_FILE_ENTRY = re.compile(
    RE_PRIORITY_AUDIT_DENY
    + r'(?P<owner>owner\s+)?'  # optionally: <owner>
    + '('
        + '(?P<bare_file>file)'  # bare 'file,' # noqa: E131
    + '|'  # or
        + r'(?P<file_keyword>file\s+)?'  # optional 'file' keyword
        + '('
           + RE_PROFILE_PATH_OR_VAR % 'path' + r'\s+' + RE_PATH_PERMS % 'perms'  # path and perms
        + '|'  # or
           + RE_PATH_PERMS % 'perms2' + r'\s+' + RE_PROFILE_PATH_OR_VAR % 'path2'  # perms and path
        + ')'
        + r'(\s+->\s*' + RE_PROFILE_NAME % 'target' + ')?'
    + '|'  # or
        + r'(?P<link_keyword>link\s+)'  # 'link' keyword
        + r'(?P<subset_keyword>subset\s+)?'  # optional 'subset' keyword
        + RE_PROFILE_PATH_OR_VAR % 'link_path'  # path
        + r'\s+' + '->' + r'\s+'  # ' -> '
        + RE_PROFILE_PATH_OR_VAR % 'link_target'  # path
    + ')'
    + RE_COMMA_EOL)

_aare = r'((,*[][!/\\\()&.*?@{}\w^-]|\\.)+)'
_quoted_aare = r'"((,*[][!/\\\()&.*?@{}\w\s^-]|\\.)+)"'  # may contain \s

aare = '(' + _aare + '|' + _quoted_aare + r'|\((' + _aare + '|' + _quoted_aare + r')\))'
aare_set = '(' + _aare + '|' + _quoted_aare + r'|\((' + _aare + '|' + _quoted_aare + r')+\))'


def re_cond_set(x, y=None):
    return r'\s*(' + x + r'\s*=\s*(?P<' + (y or x) + '_cond_set>' + aare_set + r'))[,\s]*'


def re_cond(x, y=None):
    return r'\s*(' + x + r'\s*=\s*(?P<' + (y or x) + '_cond>' + aare + r'))[,\s]*'


def parse_profile_start_line(line, filename):
    common_sections = ['leadingspace', 'flags', 'comment']

    sections = ['plainprofile', 'namedprofile', 'attachment', 'xattrs'] + common_sections
    matches = RE_PROFILE_START.search(line)

    if not matches:
        sections = ['hat_keyword', 'hat'] + common_sections
        matches = RE_PROFILE_HAT_DEF.search(line)

    if not matches:
        raise AppArmorBug('The given line from file %(filename)s is not the start of a profile: %(line)s'
                          % {'filename': filename, 'line': line})

    result = {}

    for section in sections:
        if matches.group(section):
            result[section] = matches.group(section)

            # sections with optional quotes
            if section in ('plainprofile', 'namedprofile', 'attachment', 'hat'):
                result[section] = strip_quotes(result[section])
        else:
            result[section] = None

    if result['flags'] and not result['flags'].strip():
        raise AppArmorException(
            _('Invalid syntax in %(filename)s: Empty set of flags in line %(line)s.'
              % {'filename': filename, 'line': line}))

    result['is_hat'] = False
    if result.get('hat'):
        result['is_hat'] = True
        result['profile'] = result['hat']
        if result['hat_keyword'] == '^':
            result['hat_keyword'] = False
        else:
            result['hat_keyword'] = True
        result['profile_keyword'] = True
    elif result['plainprofile']:
        result['profile'] = result['plainprofile']
        result['profile_keyword'] = False
    else:
        result['profile'] = result['namedprofile']
        result['profile_keyword'] = True
    if 'xattrs' in result:
        result['xattrs'] = re_parse_dict(result['xattrs'])
    else:
        result['xattrs'] = {}

    return result


RE_MAGIC_OR_QUOTED_PATH = '(<(?P<magicpath>.*)>|"(?P<quotedpath>.*)"|(?P<unquotedpath>[^<>"]*))'
RE_ABI = re.compile(r'^\s*#?abi\s*' + RE_MAGIC_OR_QUOTED_PATH + RE_COMMA_EOL)
RE_INCLUDE = re.compile(r'^\s*#?include(?P<ifexists>\s+if\s+exists)?\s*' + RE_MAGIC_OR_QUOTED_PATH + RE_EOL)


def re_match_include_parse(line, rule_name):
    """Matches the path for include, include if exists and abi rules

    rule_name can be 'include' or 'abi'

    Returns a tuple with
    - if the "if exists" condition is given
    - the include/abi path
    - if the path is a magic path (enclosed in <...>)
    """

    if rule_name == 'include':
        matches = RE_INCLUDE.search(line)
    elif rule_name == 'abi':
        matches = RE_ABI.search(line)
    else:
        raise AppArmorBug('re_match_include_parse() called with invalid rule name %s' % rule_name)

    if not matches:
        return None, None, None

    path = None
    ismagic = False
    if matches.group('magicpath'):
        path = matches.group('magicpath').strip()
        ismagic = True
    elif matches.group('unquotedpath'):
        path = matches.group('unquotedpath').strip()
        if re.search(r'\s', path):
            raise AppArmorException(_('Syntax error: %s must use quoted path or <...>') % rule_name)
        # LP: #1738879 - parser doesn't handle unquoted paths everywhere
        if rule_name == 'include':
            raise AppArmorException(_('Syntax error: %s must use quoted path or <...>') % rule_name)
    elif matches.group('quotedpath'):
        path = matches.group('quotedpath')
        # LP: 1738880 - parser doesn't handle relative paths everywhere, and
        # neither do we (see aa.py)
        if rule_name == 'include' and path and not path.startswith('/'):
            raise AppArmorException(_('Syntax error: %s must use quoted path or <...>') % rule_name)

    if not path:
        raise AppArmorException(_('Syntax error: %s rule with empty filename') % rule_name)

    # LP: #1738877 - parser doesn't handle files with spaces in the name
    if rule_name == 'include' and re.search(r'\s', path):
        raise AppArmorException(_('Syntax error: %s rule filename cannot contain spaces') % rule_name)

    ifexists = False
    if rule_name == 'include' and matches.group('ifexists'):
        ifexists = True

    return path, ifexists, ismagic


def re_match_include(line):
    """return path of a 'include' rule"""
    (path, ifexists, ismagic) = re_match_include_parse(line, 'include')

    if not ifexists:
        return path

    return None


def re_parse_dict(raw):
    """returns a dict where entries are comma or space separated"""
    result = {}
    if not raw:
        return result

    for key, value in re.findall(RE_DICT_ENTRY, raw):
        if value == '':
            value = None
        result[key] = value

    return result


def re_print_dict(d):
    parts = []
    for k, v in sorted(d.items()):
        if v:
            parts.append("{}={}".format(k, v))
        else:
            parts.append(k)
    return " ".join(parts)


def strip_parenthesis(data):
    """strips parenthesis from the given string and returns the strip()ped result.
       The parenthesis must be the first and last char, otherwise they won't be removed.
       Even if no parenthesis get removed, the result will be strip()ped.
    """
    if data.startswith('(') and data.endswith(')'):
        return data[1:-1].strip()
    else:
        return data.strip()


def strip_quotes(data):
    if len(data) > 1 and data.startswith('"') and data.endswith('"'):
        return data[1:-1]
    else:
        return data


def strip_braces(data):
    """strips braces from the given string and returns the strip()ped result.
       The braces must be the first and last char, otherwise they won't be removed.
       Even if no braces get removed, the result will be strip()ped.
    """
    if data.startswith('{') and data.endswith('}'):
        return data[1:-1]
    else:
        return data


def expand_var(var, var_dict, seen_vars):
    if var in seen_vars:
        raise AppArmorException(_('Circular dependency detected for variable {}').format(var))

    if var not in var_dict:
        raise AppArmorException(_('Trying to reference non-existing variable {}').format(var))

    resolved = []
    for val in var_dict[var]:
        resolved.extend(expand_string(val, var_dict, seen_vars | {var}))
    return resolved


def expand_string(s, var_dict, seen_vars):

    matches = list(RE_VARIABLE.finditer(s))
    if not matches:
        return [s]

    parts = []
    last_idx = 0
    for match in matches:
        start, end = match.span()
        if start > last_idx:
            parts.append([s[last_idx:start]])

        var_name = match.group(0)
        parts.append(expand_var(var_name, var_dict, seen_vars))
        last_idx = end

    if last_idx < len(s):
        parts.append([s[last_idx:]])

    return [''.join(p) for p in itertools.product(*parts)]


def resolve_variables(s, var_dict):
    return expand_string(s, var_dict, set())


def expand_path_braces(pattern, max_path=1000):
    """Expand only brace alternations containing a '/'; these change the path
    depth (e.g. @{bin}=/{,usr/}bin) so a single glob cannot express them.
    Within-segment braces are left untouched (the caller turns them into a glob
    wildcard or regex). Returns a set of path-shape variants."""
    i = pattern.find('{')
    while i != -1:
        # Find the '}' that closes the '{' at index i.
        # When the loop breaks, j holds the index of that matching '}'.
        depth = 0
        for j in range(i, len(pattern)):
            if pattern[j] == '{':
                depth += 1
            elif pattern[j] == '}':
                depth -= 1
                if depth == 0:
                    break
        else:
            break  # unbalanced; leave the rest untouched

        # Split the brace body pattern[i+1:j] on its top-level commas into the
        # list of alternatives, remembering whether any of them contains a '/'.
        # TODO: '\,' is wrongly treated as a separator.
        alts = []
        cur = ''
        nested = 0  # depth of braces nested inside this pattern
        has_slash = False
        for ch in pattern[i + 1:j]:
            if ch == ',' and nested == 0:
                alts.append(cur)
                cur = ''
            else:
                if ch == '{':
                    nested += 1
                elif ch == '}':
                    nested -= 1
                elif ch == '/':
                    has_slash = True
                cur += ch
        alts.append(cur)

        if has_slash:
            # A '/' in an alternative can change the path depth: recurse on it.
            variants = set()
            for alt in alts:
                variants |= expand_path_braces(pattern[:i] + alt + pattern[j + 1:], max_path)
                if len(variants) > max_path:
                    raise AppArmorException(_('Pattern %(pattern)s expands to more than %(max)d path variants. Please clean it up.') % {'pattern': pattern, 'max': max_path})
            return variants
        i = pattern.find('{', j + 1)  # within-segment brace: skip, look further
    return {pattern}



# ----------------------------------------------------------------------
#    Copyright (C) 2016 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

from apparmor.aare import AARE
from apparmor.common import AppArmorBug, AppArmorException
from apparmor.regex import RE_PROFILE_FILE_ENTRY, strip_quotes
from apparmor.rule import (
    BaseRule, BaseRuleset, check_and_split_list, logprof_value_or_all,
    parse_modifiers, quote_if_needed)
from apparmor.translations import init_translation

_ = init_translation()

allow_exec_transitions = ('ix', 'ux', 'Ux', 'px', 'Px', 'cx', 'Cx')  # 2 chars - len relevant for split_perms()
allow_exec_fallback_transitions = ('pix', 'Pix', 'cix', 'Cix', 'pux', 'PUx', 'cux', 'CUx')  # 3 chars - len relevant for split_perms()
deny_exec_transitions = ('x')
file_permissions = ('m', 'r', 'w', 'a', 'l', 'k', 'link', 'subset')  # also defines the write order
implicit_all_permissions = ('m', 'r', 'w', 'l', 'k')


class FileRule(BaseRule):
    """Class to handle and store a single file rule"""

    # Nothing external should reference this class, all external users
    # should reference the class field FileRule.ALL
    class __FileAll:
        pass

    class __FileAnyExec:
        pass

    ALL = __FileAll
    ANY_EXEC = __FileAnyExec

    rule_name = 'file'
    _match_re = RE_PROFILE_FILE_ENTRY

    def __init__(self, path, perms, exec_perms, target, owner, file_keyword=False, leading_perms=False,
                 audit=False, deny=False, allow_keyword=False, comment='', log_event=None, priority=None):
        """Initialize object

           Parameters:
           - path: string, AARE or FileRule.ALL
           - perms: string, set of chars or FileRule.ALL (must not contain exec mode)
           - exec_perms: None or string
           - target: string, AARE or FileRule.ALL
           - owner: bool
           - file_keyword: bool
           - leading_perms: bool
        """

        super().__init__(audit=audit, deny=deny, allow_keyword=allow_keyword,
                         comment=comment, log_event=log_event, priority=priority)

        #                                              rulepart  partperms  is_path  log_event
        self.path,   self.all_paths   = self._aare_or_all(path,   'path',   True,  log_event)  # noqa: E221
        self.target, self.all_targets = self._aare_or_all(target, 'target', False, log_event)

        self.can_glob = not self.all_paths
        self.can_glob_ext = not self.all_paths
        self.can_edit = not self.all_paths

        if isinstance(perms, str):
            perms, tmp_exec_perms = split_perms(perms, deny)
            if tmp_exec_perms:
                raise AppArmorBug('perms must not contain exec perms')
        elif perms is None:
            perms = set()

        if perms == {'subset'}:
            raise AppArmorBug('subset without link permissions given')
        elif perms in ({'link'}, {'link', 'subset'}):
            self.perms = perms
            self.all_perms = False
        else:
            self.perms, self.all_perms, unknown_items = check_and_split_list(
                perms, file_permissions, self.ALL, type(self).__name__, 'permissions', allow_empty_list=True)
            if unknown_items:
                raise AppArmorBug('Passed unknown perms to %s: %s' % (type(self).__name__, str(unknown_items)))
            if self.perms and 'a' in self.perms and 'w' in self.perms:
                raise AppArmorException("Conflicting permissions found: 'a' and 'w'")

        self.original_perms = None  # might be set by aa-logprof / aa.py propose_file_rules()

        if exec_perms is None:
            self.exec_perms = None
        elif 'link' in self.perms:
            raise AppArmorBug("link rules can't have execute permissions")
        elif exec_perms == self.ANY_EXEC:
            self.exec_perms = exec_perms
        elif isinstance(exec_perms, str):
            if deny:
                if exec_perms != 'x':
                    raise AppArmorException(_("file deny rules only allow to use 'x' as execute mode, but not %s" % exec_perms))
            else:
                if exec_perms == 'x':
                    raise AppArmorException(_("Execute flag ('x') in file rule must specify the exec mode (ix, Px, Cx etc.)"))
                elif exec_perms not in allow_exec_transitions and exec_perms not in allow_exec_fallback_transitions:
                    raise AppArmorBug('Unknown execute mode specified in file rule: %s' % exec_perms)
            self.exec_perms = exec_perms
        else:
            raise AppArmorBug('Passed unknown perms object to %s: %s' % (type(self).__name__, str(perms)))

        if not isinstance(owner, bool):
            raise AppArmorBug('non-boolean value passed to owner flag')
        self.owner = owner
        self.can_owner = owner  # offer '(O)wner permissions on/off' buttons only if the rule has the owner flag

        if not isinstance(file_keyword, bool):
            raise AppArmorBug('non-boolean value passed to file keyword flag')
        self.file_keyword = file_keyword

        if not isinstance(leading_perms, bool):
            raise AppArmorBug('non-boolean value passed to leading permissions flag')
        self.leading_perms = leading_perms

        # XXX subset

        # check for invalid combinations (bare 'file,' vs. path rule)
#       if (self.all_paths and not self.all_perms) or (not self.all_paths and self.all_perms):
#           raise AppArmorBug('all_paths and all_perms must be equal')
# elif
        if self.all_paths and (self.exec_perms or self.target):
            raise AppArmorBug('exec perms or target specified for bare file rule')

    @classmethod
    def _create_instance(cls, raw_rule, matches):
        """parse raw_rule and return instance of this class"""

        priority, audit, deny, allow_keyword, comment = parse_modifiers(matches)

        owner = bool(matches.group('owner'))

        leading_perms = False

        if matches.group('path'):
            path = strip_quotes(matches.group('path'))
        elif matches.group('path2'):
            path = strip_quotes(matches.group('path2'))
            leading_perms = True
        elif matches.group('link_path'):
            path = strip_quotes(matches.group('link_path'))
            leading_perms = True
        else:
            path = cls.ALL

        if matches.group('perms'):
            perms = matches.group('perms')
            perms, exec_perms = split_perms(perms, deny)
        elif matches.group('perms2'):
            perms = matches.group('perms2')
            perms, exec_perms = split_perms(perms, deny)
            leading_perms = True
        elif matches.group('link_keyword'):
            if matches.group('subset_keyword'):
                perms = {'link', 'subset'}
            else:
                perms = {'link'}
            exec_perms = None
            leading_perms = True
        else:
            perms = cls.ALL
            exec_perms = None

        if matches.group('target'):
            target = strip_quotes(matches.group('target'))
        elif matches.group('link_target'):
            target = strip_quotes(matches.group('link_target'))
        else:
            target = cls.ALL

        file_keyword = bool(matches.group('file_keyword'))

        return cls(path, perms, exec_perms, target, owner, file_keyword, leading_perms,
                   audit=audit, deny=deny, allow_keyword=allow_keyword, comment=comment, priority=priority)

    def get_clean(self, depth=0):
        """return rule (in clean/default formatting)"""

        space = '  ' * depth

        if self.all_paths:
            path = ''
        elif self.path:
            path = quote_if_needed(self.path.regex)
        else:
            raise AppArmorBug('Empty path in file rule')

        if self.all_perms:
            perms = ''
        else:
            perms = self._joint_perms()
            if not perms:
                raise AppArmorBug('Empty permissions in file rule')

        if self.leading_perms:
            path_and_perms = '%s %s' % (perms, path)
        else:
            path_and_perms = '%s %s' % (path, perms)

        if self.all_targets:
            target = ''
        elif self.target:
            target = ' -> %s' % quote_if_needed(self.target.regex)
        else:
            raise AppArmorBug('Empty exec target in file rule')

        if self.owner:
            owner = 'owner '
        else:
            owner = ''

        if self.file_keyword:
            file_keyword = 'file '
        else:
            file_keyword = ''

        if self.all_paths and self.all_perms and not path and not perms and not target:
            return ('%s%s%sfile,%s' % (space, self.modifiers_str(), owner, self.comment))  # plain 'file,' rule
        elif not self.all_paths and not self.all_perms and path and perms:
            return ('%s%s%s%s%s%s,%s' % (space, self.modifiers_str(), owner, file_keyword, path_and_perms, target, self.comment))
        else:
            raise AppArmorBug('Invalid combination of path and perms in file rule - either specify path and perms, or none of them')

    def _joint_perms(self):
        """return the permissions as string (using self.perms and self.exec_perms)"""
        return self._join_given_perms(self.perms, self.exec_perms)

    def _join_given_perms(self, perms, exec_perms):
        """return the permissions as string (using the perms and exec_perms given as parameter)"""
        perm_string = ''
        for perm in file_permissions:
            if perm in perms:
                if perm == 'subset':
                    perm = ' subset'  # add leading space
                perm_string = perm_string + perm

        if exec_perms == self.ANY_EXEC:
            raise AppArmorBug(type(self).__name__ + ".ANY_EXEC can't be used for actual rules")
        if exec_perms:
            perm_string = perm_string + exec_perms

        return perm_string

    def _is_covered_localvars(self, other_rule):
        """check if other_rule is covered by this rule object"""

        if not self._is_covered_aare(self.path, self.all_paths, other_rule.path, other_rule.all_paths, 'path'):
            return False

        if self.perms and 'subset' in self.perms and other_rule.perms and 'subset' not in other_rule.perms:
            return False  # subset is a restriction (also, if subset is included, this means this instance is a link rule, so other file permissions can't be covered)
        elif self.perms and 'link' in self.perms and other_rule.perms and 'link' in other_rule.perms:
            pass  # skip _is_covered_list() because it would interpret 'subset' as additional permissions, not as restriction
        elif not self._is_covered_list(perms_with_a(self.perms), self.all_perms,  perms_with_a(other_rule.perms), other_rule.all_perms,   'perms', sanity_check=False):
            # perms can be empty if only exec_perms are specified, therefore disable the sanity check in _is_covered_list()...
            # 'w' covers 'a', therefore use perms_with_a() to temporarily add 'a' if 'w' is present
            return False

        # TODO: check  link / link subset vs. 'l'?

        # ... and do our own sanity check
        if not other_rule.perms and not other_rule.all_perms and not other_rule.exec_perms:
            raise AppArmorBug('No permission or exec permission specified in other file rule')

        if not self.exec_perms and other_rule.exec_perms:
            return False

        # TODO: handle fallback modes?
        if other_rule.exec_perms == self.ANY_EXEC and self.exec_perms:
            pass  # other_rule has ANY_EXEC and self has an exec rule set -> covered, so avoid hitting the 'elif' branch
        elif other_rule.exec_perms and self.exec_perms != other_rule.exec_perms:
            return False

        # check exec_mode and target only if other_rule contains exec_perms (except ANY_EXEC) or link permissions
        # (for mrwk permissions, the target is ignored anyway)
        if ((other_rule.exec_perms and other_rule.exec_perms != self.ANY_EXEC)
                or (other_rule.perms and 'l' in other_rule.perms)
                or (other_rule.perms and 'link' in other_rule.perms)):
            if not self._is_covered_aare(self.target, self.all_targets, other_rule.target, other_rule.all_targets, 'target'):
                return False

            # a different target means running with a different profile, therefore we have to be more strict than _is_covered_aare()
            # XXX should we enforce an exact match for a) exec and/or b) link target?
            if self.all_targets != other_rule.all_targets:
                return False

        if self.owner and not other_rule.owner:
            return False

        # no check for file_keyword and leading_perms - they are not relevant for is_covered()

        # still here? -> then it is covered
        return True

    def _is_equal_localvars(self, other_rule, strict):
        """compare if rule-specific variables are equal"""

        if self.owner != other_rule.owner:
            return False

        if not self._is_equal_aare(self.path, self.all_paths, other_rule.path, other_rule.all_paths, 'path'):
            return False

        if self.perms != other_rule.perms:
            return False

        if self.all_perms != other_rule.all_perms:
            return False

        if self.exec_perms != other_rule.exec_perms:
            return False

        if not self._is_equal_aare(self.target, self.all_targets, other_rule.target, other_rule.all_targets, 'target'):
            return False

        if strict:  # file_keyword and leading_perms are only cosmetics, but still a difference
            if self.file_keyword != other_rule.file_keyword:
                return False

            if self.leading_perms != other_rule.leading_perms:
                return False

        return True

    def severity(self, sev_db):
        if self.all_paths:
            severity = sev_db.rank_path('/**', 'mrwlkix')
        else:
            severity = -1
            # TODO: special handling for link / link subset?
            sev = sev_db.rank_path(self.path.regex, self._joint_perms())
            if isinstance(sev, int):  # type check avoids breakage caused by 'unknown'
                severity = max(severity, sev)

        if severity == -1:
            severity = sev  # effectively 'unknown'

        return severity

    def _logprof_header_localvars(self):
        headers = []

        path = logprof_value_or_all(self.path, self.all_paths)
        headers.extend((_('Path'), path))

        old_mode = ''
        if self.original_perms:
            original_perms_set = {}
            for who in ['all', 'owner']:
                original_perms_set[who] = {}
                original_perms_set[who]['perms'] = self.original_perms['allow'][who]
                original_perms_set[who]['exec_perms'] = None

                if self.original_perms['allow'][who] == FileRule.ALL:
                    original_perms_set[who]['perms'] = set(implicit_all_permissions)
                    original_perms_set[who]['exec_perms'] = 'ix'

            original_perms_all = self._join_given_perms(original_perms_set['all']['perms'],
                                                        original_perms_set['all']['exec_perms'])
            original_perms_owner = self._join_given_perms(
                original_perms_set['owner']['perms'] - original_perms_set['all']['perms'],  # only list owner perms that are not covered by other perms
                original_perms_set['owner']['exec_perms'])

            if original_perms_all and original_perms_owner:
                old_mode = '%s + owner %s' % (original_perms_all, original_perms_owner)
            elif original_perms_all:
                old_mode = original_perms_all
            elif original_perms_owner:
                old_mode = 'owner %s' % original_perms_owner
            else:
                old_mode = ''

        if old_mode:
            headers.extend((_('Old Mode'), old_mode))

        perms = logprof_value_or_all(self.perms, self.all_perms)
        if self.perms or self.exec_perms:
            perms = self._joint_perms()

        if self.owner:
            perms = 'owner %s' % perms

        if not self.all_targets:
            perms = "%s -> %s" % (perms, self.target.regex)

        headers.extend((_('New Mode'), perms))

        # TODO: different output for link rules?

        # file_keyword and leading_perms are not really relevant
        return headers

    def glob(self):
        """Change path to next possible glob"""
        if self.all_paths:
            return

        self.path = self.path.glob_path()
        self.raw_rule = None

    def glob_ext(self):
        """Change path to next possible glob with extension"""
        if self.all_paths:
            return

        self.path = self.path.glob_path_withext()
        self.raw_rule = None

    def edit_header(self):
        if self.all_paths:
            raise AppArmorBug('Attemp to edit bare file rule')

        return (_('Enter new path: '), self.path.regex)

    def validate_edit(self, newpath):
        if self.all_paths:
            raise AppArmorBug('Attemp to edit bare file rule')

        newpath = AARE(newpath, True)  # might raise AppArmorException if the new path doesn't start with / or a variable
        return newpath.match(self.path)

    def store_edit(self, newpath):
        if self.all_paths:
            raise AppArmorBug('Attemp to edit bare file rule')

        self.path = AARE(newpath, True)  # might raise AppArmorException if the new path doesn't start with / or a variable
        self.raw_rule = None

    @staticmethod
    def hashlog_from_event(hl, ev):
        # FileRule can be of two types, "file" or "exec"
        if ev['operation'] == 'exec':
            if not ev['name']:
                raise AppArmorException('exec without executed binary')

            if not ev['name2']:
                ev['name2'] = ''  # exec events in enforce mode don't have target=...

            hl[ev['name']][ev['name2']] = True
            return

        if not ev['name']:  # missing filename -> ignore event
            return

        # Map c (create) and d (delete) to w (logging is more detailed than the profile language)
        dmask = ev['denied_mask']
        dmask = dmask.replace('c', 'w')
        dmask = dmask.replace('d', 'w')

        owner = False

        if '::' in dmask:
            # old log styles used :: to indicate if permissions are meant for owner or other
            (owner_d, other_d) = dmask.split('::')
            if owner_d and other_d:
                raise AppArmorException(
                    'Found log event with both owner and other permissions. Please open a bugreport!')
            if owner_d:
                dmask = owner_d
                owner = True
            else:
                dmask = other_d

        if ev.get('ouid') is not None and ev['fsuid'] == ev['ouid']:
            # in current log style, owner permissions are indicated by a match of fsuid and ouid
            owner = True

        if 'x' in dmask and dmask != 'x':
            dmask = dmask.replace('x', '')  # if dmask contains x and another mode, drop x here - we should see a separate exec event

        for perm in dmask:
            if perm in 'mrwalk':  # intentionally not allowing 'x' here
                hl[ev['name']][owner][perm] = True
            else:
                raise AppArmorException(_('Log contains unknown mode %s') % dmask)

    @classmethod
    def from_hashlog(cls, hl):
        for h1, h2 in BaseRule.generate_rules_from_hashlog(hl, 2):
            # FileRule can be either a 'normal' or an 'exec' file rule. These rules are encoded in hashlog differently.
            if hl[h1][h2] is True:  # Exec Rule
                name = h1
                yield FileRule(name, None, FileRule.ANY_EXEC, FileRule.ALL, owner=False, log_event=True)
            else:
                path = h1
                owner = h2
                mode = set(hl[path][owner].keys())
                # logparser sums up multiple log events, so both 'a' and 'w' can be present
                if 'a' in mode and 'w' in mode:
                    mode.remove('a')
                yield cls(path, mode, None, FileRule.ALL, owner=owner, log_event=True)
                # TODO: check for existing rules with this path, and merge them into one rule


class FileRuleset(BaseRuleset):
    """Class to handle and store a collection of file rules"""

    def get_rules_for_path(self, path, audit=False, deny=False):
        """Get all rules matching the given path
           path can be str or AARE
           If audit is True, only return rules with the audit flag set.
           If deny is True, only return matching deny rules"""

        matching_rules = type(self)()
        for rule in self.rules:
            if (rule.all_paths or rule.path.match(path)) and ((not deny) or rule.deny) and ((not audit) or rule.audit):
                matching_rules.add(rule)

        return matching_rules

    def get_perms_for_path(self, path, audit=False, deny=False):
        """Get the summarized permissions of all rules matching the given path, and the list of paths involved in the calculation
           path can be str or AARE
           If audit is True, only analyze rules with the audit flag set.
           If deny is True, only analyze matching deny rules
           Returns {'allow': {'owner': set_of_perms, 'all': set_of_perms},
                    'deny':  {'owner': set_of_perms, 'all': set_of_perms},
                    'path':  involved_paths}
           Note: exec rules and exec/link target are not honored!
        """
        # XXX do we need to honor the link target?

        perms = {
            'allow': {'owner': set(), 'all': set()},
            'deny':  {'owner': set(), 'all': set()},
        }
        all_perms = {
            'allow': {'owner': False, 'all': False},
            'deny':  {'owner': False, 'all': False},
        }
        paths = set()

        matching_rules = self.get_rules_for_path(path, audit, deny)

        for rule in matching_rules.rules:
            allow_or_deny = 'allow'
            if rule.deny:
                allow_or_deny = 'deny'

            owner_or_all = 'all'
            if rule.owner:
                owner_or_all = 'owner'

            if rule.all_perms:
                all_perms[allow_or_deny][owner_or_all] = True
            elif rule.perms:
                perms[allow_or_deny][owner_or_all] = perms[allow_or_deny][owner_or_all].union(rule.perms)
                paths.add(rule.path.regex)

        allow = {}
        deny = {}
        for who in ('all', 'owner'):
            if all_perms['allow'][who]:
                allow[who] = FileRule.ALL
            else:
                allow[who] = perms['allow'][who]

            if all_perms['deny'][who]:
                deny[who] = FileRule.ALL
            else:
                deny[who] = perms['deny'][who]

        return {'allow': allow, 'deny': deny, 'paths': paths}

    def get_exec_rules_for_path(self, path, only_exact_matches=True):
        """Get all rules matching the given path that contain exec permissions
           path can be str or AARE"""

        matches = type(self)()

        for rule in self.get_rules_for_path(path).rules:
            if rule.exec_perms:
                if rule.path.is_equal(path):
                    matches.add(rule)
                elif not only_exact_matches:
                    matches.add(rule)

        return matches

    def get_exec_conflict_rules(self, oldrule):
        """check if one of the exec rules conflict with oldrule. If yes, return the conflicting rules."""

        conflictingrules = type(self)()

        if oldrule.exec_perms:
            execrules = self.get_exec_rules_for_path(oldrule.path)

            for mergerule in execrules.rules:
                if mergerule.exec_perms != oldrule.exec_perms or mergerule.target != oldrule.target:
                    conflictingrules.add(mergerule)

        return conflictingrules


def split_perms(perm_string, deny):
    """parse permission string
       - perm_string: the permission string to parse
       - deny: True if this is a deny rule
    """
    perms = set()
    exec_mode = None

    while perm_string:
        if perm_string[0] in file_permissions:
            perms.add(perm_string[0])
            perm_string = perm_string[1:]
        elif perm_string.startswith('x'):
            if not deny:
                raise AppArmorException(_("'x' must be preceded by an exec qualifier (i, P, C or U)"))
            exec_mode = 'x'
            perm_string = perm_string[1:]
        elif perm_string.startswith(allow_exec_transitions):
            if exec_mode and exec_mode != perm_string[0:2]:
                raise AppArmorException(_('conflicting execute permissions found: %s and %s' % (exec_mode, perm_string[0:2])))
            exec_mode = perm_string[0:2]
            perm_string = perm_string[2:]
        elif perm_string.startswith(allow_exec_fallback_transitions):
            if exec_mode and exec_mode != perm_string[0:3]:
                raise AppArmorException(_('conflicting execute permissions found: %s and %s' % (exec_mode, perm_string[0:3])))
            exec_mode = perm_string[0:3]
            perm_string = perm_string[3:]
        else:
            raise AppArmorException(_('permission contains unknown character(s) %s' % perm_string))

    return perms, exec_mode


def perms_with_a(perms):
    """if perms includes 'w', add 'a' perms
       - perms: the original permissions
    """
    if not perms or 'w' not in perms:
        return perms  # no need to change anything

    perms_with_a = set(perms)
    perms_with_a.add('a')

    return perms_with_a



# ----------------------------------------------------------------------
#    Copyright (C) 2013 Kshitij Gupta <kgupta8592@gmail.com>
#    Copyright (C) 2015 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

from apparmor.common import AppArmorBug, AppArmorException
from apparmor.regex import RE_PROFILE_CHANGE_PROFILE, strip_quotes
from apparmor.rule import (
    BaseRule, BaseRuleset, logprof_value_or_all, parse_modifiers, quote_if_needed)
from apparmor.translations import init_translation

_ = init_translation()


class ChangeProfileRule(BaseRule):
    """Class to handle and store a single change_profile rule"""

    # Nothing external should reference this class, all external users
    # should reference the class field ChangeProfileRule.ALL
    class __ChangeProfileAll:
        pass

    ALL = __ChangeProfileAll

    rule_name = 'change_profile'
    equiv_execmodes = ['safe', '', None]
    _match_re = RE_PROFILE_CHANGE_PROFILE

    def __init__(self, execmode, execcond, targetprofile, audit=False, deny=False, allow_keyword=False,
                 comment='', log_event=None, priority=None):
        """CHANGE_PROFILE RULE = 'change_profile' [ [ EXEC MODE ] EXEC COND ] [ -> PROGRAMCHILD ]"""

        super().__init__(audit=audit, deny=deny, allow_keyword=allow_keyword,
                         comment=comment, log_event=log_event, priority=priority)

        if execmode:
            if execmode != 'safe' and execmode != 'unsafe':
                raise AppArmorBug('Unknown exec mode (%s) in change_profile rule' % execmode)
            elif not execcond or execcond == self.ALL:
                raise AppArmorException('Exec condition is required when unsafe or safe keywords are present')
        self.execmode = execmode

        self.execcond = None
        self.all_execconds = False
        if execcond == self.ALL:
            self.all_execconds = True
        elif isinstance(execcond, str):
            if not execcond.strip():
                raise AppArmorBug('Empty exec condition in change_profile rule')
            elif execcond.startswith('/') or execcond.startswith('@'):
                self.execcond = execcond
            else:
                raise AppArmorException('Exec condition in change_profile rule does not start with /: %s' % str(execcond))
        else:
            raise AppArmorBug('Passed unknown object to %s: %s' % (type(self).__name__, str(execcond)))

        self.targetprofile = None
        self.all_targetprofiles = False
        if targetprofile == self.ALL:
            self.all_targetprofiles = True
        elif isinstance(targetprofile, str):
            if targetprofile.strip():
                self.targetprofile = targetprofile
            else:
                raise AppArmorBug('Empty target profile in change_profile rule')
        else:
            raise AppArmorBug('Passed unknown object to %s: %s' % (type(self).__name__, str(targetprofile)))

    @classmethod
    def _create_instance(cls, raw_rule, matches):
        """parse raw_rule and return instance of this class"""

        priority, audit, deny, allow_keyword, comment = parse_modifiers(matches)

        execmode = matches.group('execmode')

        if matches.group('execcond'):
            execcond = strip_quotes(matches.group('execcond'))
        else:
            execcond = cls.ALL

        if matches.group('targetprofile'):
            targetprofile = strip_quotes(matches.group('targetprofile'))
        else:
            targetprofile = cls.ALL

        return cls(execmode, execcond, targetprofile,
                   audit=audit, deny=deny, allow_keyword=allow_keyword, comment=comment, priority=priority)

    def get_clean(self, depth=0):
        """return rule (in clean/default formatting)"""

        space = '  ' * depth

        if self.execmode:
            execmode = ' %s' % self.execmode
        else:
            execmode = ''

        if self.all_execconds:
            execcond = ''
        elif self.execcond:
            execcond = ' %s' % quote_if_needed(self.execcond)
        else:
            raise AppArmorBug('Empty execcond in change_profile rule')

        if self.all_targetprofiles:
            targetprofile = ''
        elif self.targetprofile:
            targetprofile = ' -> %s' % quote_if_needed(self.targetprofile)
        else:
            raise AppArmorBug('Empty target profile in change_profile rule')

        return ('%s%schange_profile%s%s%s,%s' % (space, self.modifiers_str(), execmode, execcond, targetprofile, self.comment))

    def _is_covered_localvars(self, other_rule):
        """check if other_rule is covered by this rule object"""

        if (self.execmode != other_rule.execmode
                and (self.execmode not in self.equiv_execmodes
                     or other_rule.execmode not in self.equiv_execmodes)):
            return False

        if not self._is_covered_plain(self.execcond, self.all_execconds, other_rule.execcond, other_rule.all_execconds, 'exec condition'):
            # TODO: honor globbing and variables
            return False

        if not self._is_covered_plain(self.targetprofile, self.all_targetprofiles, other_rule.targetprofile, other_rule.all_targetprofiles, 'target profile'):
            return False

        # still here? -> then it is covered
        return True

    def _is_equal_localvars(self, other_rule, strict):
        """compare if rule-specific variables are equal"""

        if (self.execmode != other_rule.execmode
                and (self.execmode not in self.equiv_execmodes
                     or other_rule.execmode not in self.equiv_execmodes)):
            return False

        if (self.execcond != other_rule.execcond
                or self.all_execconds != other_rule.all_execconds):
            return False

        if (self.targetprofile != other_rule.targetprofile
                or self.all_targetprofiles != other_rule.all_targetprofiles):
            return False

        return True

    def _logprof_header_localvars(self):
        headers = []

        if self.execmode:
            headers.extend((_('Exec Mode'), self.execmode))

        execcond_txt       = logprof_value_or_all(self.execcond,      self.all_execconds)  # noqa: E221
        targetprofiles_txt = logprof_value_or_all(self.targetprofile, self.all_targetprofiles)

        headers.extend((
            _('Exec Condition'), execcond_txt,
            _('Target Profile'), targetprofiles_txt,
        ))
        return headers

    @staticmethod
    def hashlog_from_event(hl, ev):
        hl[ev['name2']] = True

    @classmethod
    def from_hashlog(cls, hl):
        for cp in hl.keys():
            if cp:  # skip events without target profile
                yield cls(None, cls.ALL, cp, log_event=True)


class ChangeProfileRuleset(BaseRuleset):
    """Class to handle and store a collection of change_profile rules"""

    def get_glob(self, path_or_rule):
        """Return the next possible glob. For change_profile rules, that can be "change_profile EXECCOND,",
           "change_profile -> TARGET_PROFILE," or "change_profile," (all change_profile).
           Also, EXECCOND filename can be globbed"""
        # XXX implement all options mentioned above ;-)
        return 'change_profile,'



# ----------------------------------------------------------------------
#    Copyright (C) 2023 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

from apparmor.regex import RE_PROFILE_ALL
from apparmor.rule import BaseRule, BaseRuleset, parse_modifiers
from apparmor.translations import init_translation

_ = init_translation()


class AllRule(BaseRule):
    """Class to handle and store a single all rule"""

    # This class doesn't have any localvars, therefore it doesn't need 'ALL'

    can_glob = False
    rule_name = 'all'
    _match_re = RE_PROFILE_ALL

    def __init__(self, audit=False, deny=False, allow_keyword=False,
                 comment='', log_event=None, priority=None):

        super().__init__(audit=audit, deny=deny, allow_keyword=allow_keyword,
                         comment=comment, log_event=log_event, priority=priority)

        # no localvars -> nothing more to do

    @classmethod
    def _create_instance(cls, raw_rule, matches):
        """parse raw_rule and return instance of this class"""

        priority, audit, deny, allow_keyword, comment = parse_modifiers(matches)

        return cls(audit=audit, deny=deny,
                   allow_keyword=allow_keyword,
                   comment=comment, priority=priority)

    def get_clean(self, depth=0):
        """return rule (in clean/default formatting)"""

        space = '  ' * depth

        return ('%s%sall,%s' % (space, self.modifiers_str(), self.comment))

    def _is_covered_localvars(self, other_rule):
        """check if other_rule is covered by this rule object"""

        # no localvars, so there can't be a difference
        return True

    def _is_equal_localvars(self, other_rule, strict):
        """compare if rule-specific variables are equal"""

        # no localvars, so there can't be a difference
        return True

    def severity(self, sev_db):
        # allowing _everything_ is the worst thing you could do, therefore hardcode highest severity
        severity = 10

        return severity

    def _logprof_header_localvars(self):
        return _('All'), _('Allow everything')


class AllRuleset(BaseRuleset):
    """Class to handle and store a collection of all rules"""

    def get_glob(self, path_or_rule):
        # There's nothing to glob in all rules
        raise NotImplementedError
